package net.minecraft.entity.boss;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.google.common.collect.Lists;

import net.lax1dude.eaglercraft.v1_8.EaglercraftRandom;
import net.minecraft.block.Block;
import net.minecraft.block.BlockTorch;
import net.minecraft.block.material.Material;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLiving;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.IEntityMultiPart;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.item.EntityEnderCrystal;
import net.minecraft.entity.item.EntityXPOrb;
import net.minecraft.entity.monster.IMob;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.util.AxisAlignedBB;
import net.minecraft.util.BlockPos;
import net.minecraft.util.DamageSource;
import net.minecraft.util.EntityDamageSource;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.MathHelper;
import net.minecraft.util.Vec3;
import net.minecraft.world.Explosion;
import net.minecraft.world.World;

/**
 * + This portion of EaglercraftX contains deobfuscated Minecraft 1.8 source
 * code.
 *
 * Minecraft 1.8.8 bytecode is (c) 2015 Mojang AB. "Do not distribute!" Mod
 * Coder Pack v9.18 deobfuscation configs are (c) Copyright by the MCP Team
 *
 * EaglercraftX 1.8 patch files (c) 2022-2025 lax1dude, ayunami2000. All Rights
 * Reserved.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
public class EntityDragon extends EntityLiving implements IBossDisplayData, IEntityMultiPart, IMob {
	public double targetX;
	public double targetY;
	public double targetZ;
	/**
	 * + Ring buffer array for the last 64 Y-positions and yaw rotations. Used to
	 * calculate offsets for the animations.
	 */
	public double[][] ringBuffer = new double[64][3];
	/**
	 * + Index into the ring buffer. Incremented once per tick and restarts at 0
	 * once it reaches the end of the buffer.
	 */
	public int ringBufferIndex = -1;
	public EntityDragonPart[] dragonPartArray;
	public EntityDragonPart dragonPartHead;
	public EntityDragonPart dragonPartBody;
	public EntityDragonPart dragonPartTail1;
	public EntityDragonPart dragonPartTail2;
	public EntityDragonPart dragonPartTail3;
	public EntityDragonPart dragonPartWing1;
	public EntityDragonPart dragonPartWing2;
	public float prevAnimTime;
	public float animTime;
	public boolean forceNewTarget;
	public boolean slowed;
	private Entity target;
	public int deathTicks;
	public EntityEnderCrystal healingEnderCrystal;

	public EntityDragon(World worldIn) {
		super(worldIn);
		this.dragonPartArray = new EntityDragonPart[] {
				this.dragonPartHead = new EntityDragonPart(this, "head", 6.0F, 6.0F),
				this.dragonPartBody = new EntityDragonPart(this, "body", 8.0F, 8.0F),
				this.dragonPartTail1 = new EntityDragonPart(this, "tail", 4.0F, 4.0F),
				this.dragonPartTail2 = new EntityDragonPart(this, "tail", 4.0F, 4.0F),
				this.dragonPartTail3 = new EntityDragonPart(this, "tail", 4.0F, 4.0F),
				this.dragonPartWing1 = new EntityDragonPart(this, "wing", 4.0F, 4.0F),
				this.dragonPartWing2 = new EntityDragonPart(this, "wing", 4.0F, 4.0F) };
		this.setHealth(this.getMaxHealth());
		this.setSize(16.0F, 8.0F);
		this.noClip = true;
		this.isImmuneToFire = true;
		this.targetY = 100.0D;
		this.ignoreFrustumCheck = true;
	}

	@Override
	protected void applyEntityAttributes() {
		super.applyEntityAttributes();
		this.getEntityAttribute(SharedMonsterAttributes.maxHealth).setBaseValue(200.0D);
	}

	/**
	 * + Provides a way to cause damage to an ender dragon.
	 */
	protected boolean attackDragonFrom(DamageSource source, float amount) {
		return super.attackEntityFrom(source, amount);
	}

	/**
	 * + Attacks all entities inside this list, dealing 5 hearts of damage.
	 */
	private void attackEntitiesInList(List<Entity> parList) {
		for (int i = 0, l = parList.size(); i < l; ++i) {
			Entity entity = (Entity) parList.get(i);
			if (entity instanceof EntityLivingBase) {
				entity.attackEntityFrom(DamageSource.causeMobDamage(this), 10.0F);
				this.applyEnchantments(this, entity);
			}
		}

	}

	/**
	 * + Called when the entity is attacked.
	 */
	@Override
	public boolean attackEntityFrom(DamageSource damagesource, float f) {
		if (damagesource instanceof EntityDamageSource && ((EntityDamageSource) damagesource).getIsThornsDamage()) {
			this.attackDragonFrom(damagesource, f);
		}

		return false;
	}

	@Override
	public boolean attackEntityFromPart(EntityDragonPart entitydragonpart, DamageSource damagesource, float f) {
		if (entitydragonpart != this.dragonPartHead) {
			f = f / 4.0F + 1.0F;
		}

		float f1 = this.rotationYaw * 3.1415927F / 180.0F;
		float f2 = MathHelper.sin(f1);
		float f3 = MathHelper.cos(f1);
		this.targetX = this.posX + (double) (f2 * 5.0F) + (double) ((this.rand.nextFloat() - 0.5F) * 2.0F);
		this.targetY = this.posY + (double) (this.rand.nextFloat() * 3.0F) + 1.0D;
		this.targetZ = this.posZ - (double) (f3 * 5.0F) + (double) ((this.rand.nextFloat() - 0.5F) * 2.0F);
		this.target = null;
		if (damagesource.getEntity() instanceof EntityPlayer || damagesource.isExplosion()) {
			this.attackDragonFrom(damagesource, f);
		}

		return true;
	}

	/**
	 * + Returns true if other Entities should be prevented from moving through this
	 * Entity.
	 */
	@Override
	public boolean canBeCollidedWith() {
		return false;
	}

	/**
	 * + Pushes all entities inside the list away from the enderdragon.
	 */
	private void collideWithEntities(List<Entity> parList) {
		double d0 = (this.dragonPartBody.getEntityBoundingBox().minX + this.dragonPartBody.getEntityBoundingBox().maxX)
				/ 2.0D;
		double d1 = (this.dragonPartBody.getEntityBoundingBox().minZ + this.dragonPartBody.getEntityBoundingBox().maxZ)
				/ 2.0D;

		for (int i = 0, l = parList.size(); i < l; ++i) {
			Entity entity = parList.get(i);
			if (entity instanceof EntityLivingBase) {
				double d2 = entity.posX - d0;
				double d3 = entity.posZ - d1;
				double d4 = d2 * d2 + d3 * d3;
				entity.addVelocity(d2 / d4 * 4.0D, 0.20000000298023224D, d3 / d4 * 4.0D);
			}
		}

	}

	private void createCrystal(BlockPos basePos) {
		int height = 8;
		int multiplier = 2;
		int maxHeight = height + 1;
		for (int i = 0; i < height; i++) {
			int width = (height - i - 1) * multiplier / 2;
			for (int j = -width; j <= width; j++) {
				for (int k = -width; k <= width; k++) {
					if (Math.abs(j) + Math.abs(k) <= width) {
						BlockPos pos = basePos.add(j, i - 1, k);
						this.worldObj.setBlockState(pos, getCrystalRandomState());
					}
				}
			}
		}
		BlockPos reverseBasePos = basePos.add(0, -1, 0);
		this.worldObj.setBlockState(reverseBasePos, Blocks.dragon_egg.getDefaultState());

		reverseBasePos = basePos.add(0, -1, 0);

		for (int i = 0; i < height; i++) {
			int width = (height - i - 1) * multiplier / 2;
			for (int j = -width; j <= width; j++) {
				for (int k = -width; k <= width; k++) {
					if (Math.abs(j) + Math.abs(k) <= width) {
						BlockPos pos = reverseBasePos.add(j, -i - 1, k);
						this.worldObj.setBlockState(pos, getCrystalRandomState());
					}
				}
			}
		}
	}

	/**
	 * + Makes the entity despawn if requirements are reached
	 */
	@Override
	protected void despawnEntity() {
	}

	/**
	 * + Destroys all blocks that aren't associated with 'The End' inside the given
	 * bounding box.
	 */
	private boolean destroyBlocksInAABB(AxisAlignedBB parAxisAlignedBB) {
		int i = MathHelper.floor_double(parAxisAlignedBB.minX);
		int j = MathHelper.floor_double(parAxisAlignedBB.minY);
		int k = MathHelper.floor_double(parAxisAlignedBB.minZ);
		int l = MathHelper.floor_double(parAxisAlignedBB.maxX);
		int i1 = MathHelper.floor_double(parAxisAlignedBB.maxY);
		int j1 = MathHelper.floor_double(parAxisAlignedBB.maxZ);
		boolean flag = false;
		boolean flag1 = false;

		for (int k1 = i; k1 <= l; ++k1) {
			for (int l1 = j; l1 <= i1; ++l1) {
				for (int i2 = k; i2 <= j1; ++i2) {
					BlockPos blockpos = new BlockPos(k1, l1, i2);
					Block block = this.worldObj.getBlockState(blockpos).getBlock();
					if (block.getMaterial() != Material.air) {
						if (block != Blocks.barrier && block != Blocks.obsidian && block != Blocks.end_stone
								&& block != Blocks.bedrock && block != Blocks.command_block
								&& block != Blocks.dragonite_block
								&& this.worldObj.getGameRules().getBoolean("mobGriefing")) {
							flag1 = this.worldObj.setBlockToAir(blockpos) || flag1;
						} else {
							flag = true;
						}
					}
				}
			}
		}

		if (flag1) {
			double d0 = parAxisAlignedBB.minX
					+ (parAxisAlignedBB.maxX - parAxisAlignedBB.minX) * (double) this.rand.nextFloat();
			double d1 = parAxisAlignedBB.minY
					+ (parAxisAlignedBB.maxY - parAxisAlignedBB.minY) * (double) this.rand.nextFloat();
			double d2 = parAxisAlignedBB.minZ
					+ (parAxisAlignedBB.maxZ - parAxisAlignedBB.minZ) * (double) this.rand.nextFloat();
			this.worldObj.spawnParticle(EnumParticleTypes.EXPLOSION_LARGE, d0, d1, d2, 0.0D, 0.0D, 0.0D, new int[0]);
		}

		return flag;
	}

	@Override
	protected void entityInit() {
		super.entityInit();
	}

	/**
	 * + Generate the portal when the dragon dies
	 */
	private void generatePortal() {
		BlockPos surfacePos = this.worldObj.getTopSolidOrLiquidBlock(new BlockPos(0, 30, 0));
		BlockPos pos = surfacePos.up();

		System.out.println("genportal");

		for (int i = -1; i <= 32; ++i) {
			for (int j = -5; j <= 5; ++j) {
				for (int k = -5; k <= 5; ++k) {
					double d2 = (double) (j * j + k * k);
					BlockPos blockpos = pos.add(j, i, k);

					if (d2 <= 16.0D) {
						if (i < 0) {
							if (d2 <= 9.0D) {
								this.worldObj.setBlockState(blockpos, Blocks.obsidian.getDefaultState());
							}
						} else if (i > 0) {
							this.worldObj.setBlockState(blockpos, Blocks.air.getDefaultState());
						} else if (d2 > 9.0D) {
							this.worldObj.setBlockState(blockpos, Blocks.obsidian.getDefaultState());
						} else {
							this.worldObj.setBlockState(blockpos, Blocks.end_portal.getDefaultState());
						}
					}
				}
			}
		}

		this.worldObj.setBlockState(pos, Blocks.end_stone.getDefaultState());
		this.worldObj.setBlockState(pos.up(), Blocks.end_stone.getDefaultState());
		this.worldObj.setBlockState(pos.up().up(), Blocks.dragon_egg.getDefaultState());

		BlockPos blockpos1 = pos.up(2);
		this.worldObj.setBlockState(blockpos1, Blocks.end_stone.getDefaultState());
		this.worldObj.setBlockState(blockpos1.west(),
				Blocks.torch.getDefaultState().withProperty(BlockTorch.FACING, EnumFacing.EAST));
		this.worldObj.setBlockState(blockpos1.east(),
				Blocks.torch.getDefaultState().withProperty(BlockTorch.FACING, EnumFacing.WEST));
		this.worldObj.setBlockState(blockpos1.north(),
				Blocks.torch.getDefaultState().withProperty(BlockTorch.FACING, EnumFacing.SOUTH));
		this.worldObj.setBlockState(blockpos1.south(),
				Blocks.torch.getDefaultState().withProperty(BlockTorch.FACING, EnumFacing.NORTH));

		this.worldObj.setBlockState(pos.up(3), Blocks.end_stone.getDefaultState());

		BlockPos beaconPos = pos.up(12);

		for (int i = 1; i <= 64; ++i) {
			this.worldObj.setBlockState(beaconPos.up(i), Blocks.air.getDefaultState());
		}

		BlockPos basePos = beaconPos.down();

		this.worldObj.setBlockState(basePos.down(), Blocks.dragon_egg.getDefaultState());

		BlockPos crystalBase = beaconPos.up(7);
		createCrystal(crystalBase);
	}

	private IBlockState getCrystalRandomState() {
		EaglercraftRandom random = new EaglercraftRandom();
		int rand = random.nextInt(7);
		switch (rand) {
		case 0:
			return Blocks.stained_glass.getStateFromMeta(6);
		case 1:
			return Blocks.end_stone.getDefaultState();
		case 2:
			return Blocks.purple_packed_terracotta.getDefaultState();
		case 3:
			return Blocks.stained_glass.getStateFromMeta(6);
		case 4:
			return Blocks.end_stone.getDefaultState();
		case 5:
			return Blocks.dragonite_block.getDefaultState();
		default:
			return Blocks.end_stone.getDefaultState();
		}
	}

	/**
	 * + Returns the sound this mob makes when it is hurt.
	 */
	@Override
	protected String getHurtSound() {
		return "mob.enderdragon.hit";
	}

	/**
	 * + Returns the sound this mob makes while it's alive.
	 */
	@Override
	protected String getLivingSound() {
		return "mob.enderdragon.growl";
	}

	/**
	 * + Returns a double[3] array with movement offsets, used to calculate trailing
	 * tail/neck positions. [0] = yaw offset, [1] = y offset, [2] = unused, always
	 * 0. Parameters: buffer index offset, partial ticks.
	 */
	public double[] getMovementOffsets(int parInt1, float parFloat1) {
		if (this.getHealth() <= 0.0F) {
			parFloat1 = 0.0F;
		}

		parFloat1 = 1.0F - parFloat1;
		int i = this.ringBufferIndex - parInt1 * 1 & 63;
		int j = this.ringBufferIndex - parInt1 * 1 - 1 & 63;
		double[] adouble = new double[3];
		double d0 = this.ringBuffer[i][0];
		double d1 = MathHelper.wrapAngleTo180_double(this.ringBuffer[j][0] - d0);
		adouble[0] = d0 + d1 * (double) parFloat1;
		d0 = this.ringBuffer[i][1];
		d1 = this.ringBuffer[j][1] - d0;
		adouble[1] = d0 + d1 * (double) parFloat1;
		adouble[2] = this.ringBuffer[i][2] + (this.ringBuffer[j][2] - this.ringBuffer[i][2]) * (double) parFloat1;
		return adouble;
	}

	/**
	 * + Return the Entity parts making up this Entity (currently only for dragons)
	 */
	@Override
	public Entity[] getParts() {
		return this.dragonPartArray;
	}

	/**
	 * + Returns the volume for the sounds this mob makes.
	 */
	@Override
	protected float getSoundVolume() {
		return 5.0F;
	}

	@Override
	public World getWorld() {
		return this.worldObj;
	}

	/**
	 * + handles entity death timer, experience orb and particle creation
	 */
	@Override
	protected void onDeathUpdate() {
		++this.deathTicks;
		if (this.deathTicks >= 180 && this.deathTicks <= 200) {
			float f = (this.rand.nextFloat() - 0.5F) * 8.0F;
			float f1 = (this.rand.nextFloat() - 0.5F) * 4.0F;
			float f2 = (this.rand.nextFloat() - 0.5F) * 8.0F;
			this.worldObj.spawnParticle(EnumParticleTypes.EXPLOSION_HUGE, this.posX + (double) f,
					this.posY + 2.0D + (double) f1, this.posZ + (double) f2, 0.0D, 0.0D, 0.0D, new int[0]);
		}

		boolean flag = this.worldObj.getGameRules().getBoolean("doMobLoot");
		if (!this.worldObj.isRemote) {
			if (this.deathTicks > 150 && this.deathTicks % 5 == 0 && flag) {
				int i = 1000;

				while (i > 0) {
					int k = EntityXPOrb.getXPSplit(i);
					i -= k;
					this.worldObj
							.spawnEntityInWorld(new EntityXPOrb(this.worldObj, this.posX, this.posY, this.posZ, k));
				}
			}

			if (this.deathTicks == 1) {
				this.worldObj.playBroadcastSound(1018, new BlockPos(this), 0);
			}
		}

		this.moveEntity(0.0D, 0.10000000149011612D, 0.0D);
		this.renderYawOffset = this.rotationYaw += 20.0F;
		if (this.deathTicks == 200 && !this.worldObj.isRemote) {
			if (flag) {
				int j = 2000;

				while (j > 0) {
					int l = EntityXPOrb.getXPSplit(j);
					j -= l;
					this.worldObj
							.spawnEntityInWorld(new EntityXPOrb(this.worldObj, this.posX, this.posY, this.posZ, l));
				}
			}

			this.generatePortal();
			this.setDead();
		}

	}

	/**
	 * + Called by the /kill command.
	 */
	@Override
	public void onKillCommand() {
		this.generatePortal();
		this.setDead();
	}

	/**
	 * + Called frequently so the entity can update its state every tick as
	 * required. For example, zombies and skeletons use this to react to sunlight
	 * and start to burn.
	 */
	@Override
	public void onLivingUpdate() {
		if (this.worldObj.isRemote) {
			float f = MathHelper.cos(this.animTime * 3.1415927F * 2.0F);
			float f1 = MathHelper.cos(this.prevAnimTime * 3.1415927F * 2.0F);
			if (f1 <= -0.3F && f >= -0.3F && !this.isSilent()) {
				this.worldObj.playSound(this.posX, this.posY, this.posZ, "mob.enderdragon.wings", 5.0F,
						0.8F + this.rand.nextFloat() * 0.3F, false);
			}
		}

		this.prevAnimTime = this.animTime;
		if (this.getHealth() <= 0.0F) {
			float f11 = (this.rand.nextFloat() - 0.5F) * 8.0F;
			float f13 = (this.rand.nextFloat() - 0.5F) * 4.0F;
			float f14 = (this.rand.nextFloat() - 0.5F) * 8.0F;
			this.worldObj.spawnParticle(EnumParticleTypes.EXPLOSION_LARGE, this.posX + (double) f11,
					this.posY + 2.0D + (double) f13, this.posZ + (double) f14, 0.0D, 0.0D, 0.0D, new int[0]);
		} else {
			this.updateDragonEnderCrystal();
			float f10 = 0.2F
					/ (MathHelper.sqrt_double(this.motionX * this.motionX + this.motionZ * this.motionZ) * 10.0F
							+ 1.0F);
			f10 = f10 * (float) Math.pow(2.0D, this.motionY);
			if (this.slowed) {
				this.animTime += f10 * 0.5F;
			} else {
				this.animTime += f10;
			}

			this.rotationYaw = MathHelper.wrapAngleTo180_float(this.rotationYaw);
			if (this.isAIDisabled()) {
				this.animTime = 0.5F;
			} else {
				if (this.ringBufferIndex < 0) {
					for (int i = 0; i < this.ringBuffer.length; ++i) {
						this.ringBuffer[i][0] = (double) this.rotationYaw;
						this.ringBuffer[i][1] = this.posY;
					}
				}

				if (++this.ringBufferIndex == this.ringBuffer.length) {
					this.ringBufferIndex = 0;
				}

				this.ringBuffer[this.ringBufferIndex][0] = (double) this.rotationYaw;
				this.ringBuffer[this.ringBufferIndex][1] = this.posY;
				if (this.worldObj.isRemote) {
					if (this.newPosRotationIncrements > 0) {
						double d10 = this.posX + (this.newPosX - this.posX) / (double) this.newPosRotationIncrements;
						double d0 = this.posY + (this.newPosY - this.posY) / (double) this.newPosRotationIncrements;
						double d1 = this.posZ + (this.newPosZ - this.posZ) / (double) this.newPosRotationIncrements;
						double d2 = MathHelper.wrapAngleTo180_double(this.newRotationYaw - (double) this.rotationYaw);
						this.rotationYaw = (float) ((double) this.rotationYaw
								+ d2 / (double) this.newPosRotationIncrements);
						this.rotationPitch = (float) ((double) this.rotationPitch
								+ (this.newRotationPitch - (double) this.rotationPitch)
										/ (double) this.newPosRotationIncrements);
						--this.newPosRotationIncrements;
						this.setPosition(d10, d0, d1);
						this.setRotation(this.rotationYaw, this.rotationPitch);
					}
				} else {
					double d11 = this.targetX - this.posX;
					double d12 = this.targetY - this.posY;
					double d13 = this.targetZ - this.posZ;
					double d14 = d11 * d11 + d12 * d12 + d13 * d13;
					if (this.target != null) {
						this.targetX = this.target.posX;
						this.targetZ = this.target.posZ;
						double d3 = this.targetX - this.posX;
						double d5 = this.targetZ - this.posZ;
						double d7 = Math.sqrt(d3 * d3 + d5 * d5);
						double d8 = 0.4000000059604645D + d7 / 80.0D - 1.0D;
						if (d8 > 10.0D) {
							d8 = 10.0D;
						}

						this.targetY = this.target.getEntityBoundingBox().minY + d8;
					} else {
						this.targetX += this.rand.nextGaussian() * 2.0D;
						this.targetZ += this.rand.nextGaussian() * 2.0D;
					}

					if (this.forceNewTarget || d14 < 100.0D || d14 > 22500.0D || this.isCollidedHorizontally
							|| this.isCollidedVertically) {
						this.setNewTarget();
					}

					d12 = d12 / (double) MathHelper.sqrt_double(d11 * d11 + d13 * d13);
					float f17 = 0.6F;
					d12 = MathHelper.clamp_double(d12, (double) (-f17), (double) f17);
					this.motionY += d12 * 0.10000000149011612D;
					this.rotationYaw = MathHelper.wrapAngleTo180_float(this.rotationYaw);
					double d4 = 180.0D - MathHelper.func_181159_b(d11, d13) * 180.0D / 3.1415927410125732D;
					double d6 = MathHelper.wrapAngleTo180_double(d4 - (double) this.rotationYaw);
					if (d6 > 50.0D) {
						d6 = 50.0D;
					}

					if (d6 < -50.0D) {
						d6 = -50.0D;
					}

					Vec3 vec3 = (new Vec3(this.targetX - this.posX, this.targetY - this.posY, this.targetZ - this.posZ))
							.normalize();
					double d15 = (double) (-MathHelper.cos(this.rotationYaw * 3.1415927F / 180.0F));
					Vec3 vec31 = (new Vec3((double) MathHelper.sin(this.rotationYaw * 3.1415927F / 180.0F),
							this.motionY, d15)).normalize();
					float f5 = ((float) vec31.dotProduct(vec3) + 0.5F) / 1.5F;
					if (f5 < 0.0F) {
						f5 = 0.0F;
					}

					this.randomYawVelocity *= 0.8F;
					float f6 = MathHelper.sqrt_double(this.motionX * this.motionX + this.motionZ * this.motionZ) * 1.0F
							+ 1.0F;
					double d9 = Math.sqrt(this.motionX * this.motionX + this.motionZ * this.motionZ) * 1.0D + 1.0D;
					if (d9 > 40.0D) {
						d9 = 40.0D;
					}

					this.randomYawVelocity = (float) ((double) this.randomYawVelocity
							+ d6 * (0.699999988079071D / d9 / (double) f6));
					this.rotationYaw += this.randomYawVelocity * 0.1F;
					float f7 = (float) (2.0D / (d9 + 1.0D));
					float f8 = 0.06F;
					this.moveFlying(0.0F, -1.0F, f8 * (f5 * f7 + (1.0F - f7)));
					if (this.slowed) {
						this.moveEntity(this.motionX * 0.800000011920929D, this.motionY * 0.800000011920929D,
								this.motionZ * 0.800000011920929D);
					} else {
						this.moveEntity(this.motionX, this.motionY, this.motionZ);
					}

					Vec3 vec32 = (new Vec3(this.motionX, this.motionY, this.motionZ)).normalize();
					float f9 = ((float) vec32.dotProduct(vec31) + 1.0F) / 2.0F;
					f9 = 0.8F + 0.15F * f9;
					this.motionX *= (double) f9;
					this.motionZ *= (double) f9;
					this.motionY *= 0.9100000262260437D;
				}

				this.renderYawOffset = this.rotationYaw;
				this.dragonPartHead.width = this.dragonPartHead.height = 3.0F;
				this.dragonPartTail1.width = this.dragonPartTail1.height = 2.0F;
				this.dragonPartTail2.width = this.dragonPartTail2.height = 2.0F;
				this.dragonPartTail3.width = this.dragonPartTail3.height = 2.0F;
				this.dragonPartBody.height = 3.0F;
				this.dragonPartBody.width = 5.0F;
				this.dragonPartWing1.height = 2.0F;
				this.dragonPartWing1.width = 4.0F;
				this.dragonPartWing2.height = 3.0F;
				this.dragonPartWing2.width = 4.0F;
				float f12 = (float) (this.getMovementOffsets(5, 1.0F)[1] - this.getMovementOffsets(10, 1.0F)[1]) * 10.0F
						/ 180.0F * 3.1415927F;
				float f2 = MathHelper.cos(f12);
				float f15 = -MathHelper.sin(f12);
				float f3 = this.rotationYaw * 3.1415927F / 180.0F;
				float f16 = MathHelper.sin(f3);
				float f4 = MathHelper.cos(f3);
				this.dragonPartBody.onUpdate();
				this.dragonPartBody.setLocationAndAngles(this.posX + (double) (f16 * 0.5F), this.posY,
						this.posZ - (double) (f4 * 0.5F), 0.0F, 0.0F);
				this.dragonPartWing1.onUpdate();
				this.dragonPartWing1.setLocationAndAngles(this.posX + (double) (f4 * 4.5F), this.posY + 2.0D,
						this.posZ + (double) (f16 * 4.5F), 0.0F, 0.0F);
				this.dragonPartWing2.onUpdate();
				this.dragonPartWing2.setLocationAndAngles(this.posX - (double) (f4 * 4.5F), this.posY + 2.0D,
						this.posZ - (double) (f16 * 4.5F), 0.0F, 0.0F);
				if (!this.worldObj.isRemote && this.hurtTime == 0) {
					this.collideWithEntities(
							this.worldObj.getEntitiesWithinAABBExcludingEntity(this, this.dragonPartWing1
									.getEntityBoundingBox().expand(4.0D, 2.0D, 4.0D).offset(0.0D, -2.0D, 0.0D)));
					this.collideWithEntities(
							this.worldObj.getEntitiesWithinAABBExcludingEntity(this, this.dragonPartWing2
									.getEntityBoundingBox().expand(4.0D, 2.0D, 4.0D).offset(0.0D, -2.0D, 0.0D)));
					this.attackEntitiesInList(this.worldObj.getEntitiesWithinAABBExcludingEntity(this,
							this.dragonPartHead.getEntityBoundingBox().expand(1.0D, 1.0D, 1.0D)));
				}

				double[] adouble1 = this.getMovementOffsets(5, 1.0F);
				double[] adouble = this.getMovementOffsets(0, 1.0F);
				float f18 = MathHelper.sin(this.rotationYaw * 3.1415927F / 180.0F - this.randomYawVelocity * 0.01F);
				float f19 = MathHelper.cos(this.rotationYaw * 3.1415927F / 180.0F - this.randomYawVelocity * 0.01F);
				this.dragonPartHead.onUpdate();
				this.dragonPartHead.setLocationAndAngles(this.posX + (double) (f18 * 5.5F * f2),
						this.posY + (adouble[1] - adouble1[1]) * 1.0D + (double) (f15 * 5.5F),
						this.posZ - (double) (f19 * 5.5F * f2), 0.0F, 0.0F);

				for (int j = 0; j < 3; ++j) {
					EntityDragonPart entitydragonpart = null;
					if (j == 0) {
						entitydragonpart = this.dragonPartTail1;
					}

					if (j == 1) {
						entitydragonpart = this.dragonPartTail2;
					}

					if (j == 2) {
						entitydragonpart = this.dragonPartTail3;
					}

					double[] adouble2 = this.getMovementOffsets(12 + j * 2, 1.0F);
					float f20 = this.rotationYaw * 3.1415927F / 180.0F
							+ this.simplifyAngle(adouble2[0] - adouble1[0]) * 3.1415927F / 180.0F * 1.0F;
					float f21 = MathHelper.sin(f20);
					float f22 = MathHelper.cos(f20);
					float f23 = 1.5F;
					float f24 = (float) (j + 1) * 2.0F;
					entitydragonpart.onUpdate();
					entitydragonpart.setLocationAndAngles(this.posX - (double) ((f16 * f23 + f21 * f24) * f2),
							this.posY + (adouble2[1] - adouble1[1]) * 1.0D - (double) ((f24 + f23) * f15) + 1.5D,
							this.posZ + (double) ((f4 * f23 + f22 * f24) * f2), 0.0F, 0.0F);
				}

				if (!this.worldObj.isRemote) {
					this.slowed = this.destroyBlocksInAABB(this.dragonPartHead.getEntityBoundingBox())
							| this.destroyBlocksInAABB(this.dragonPartBody.getEntityBoundingBox());
				}

			}
		}
	}

	/**
	 * + Sets a new target for the flight AI. It can be a random coordinate or a
	 * nearby player.
	 */
	private void setNewTarget() {
		this.forceNewTarget = false;
		ArrayList arraylist = Lists.newArrayList(this.worldObj.playerEntities);
		Iterator iterator = arraylist.iterator();

		while (iterator.hasNext()) {
			if (((EntityPlayer) iterator.next()).isSpectator()) {
				iterator.remove();
			}
		}

		if (this.rand.nextInt(2) == 0 && !arraylist.isEmpty()) {
			this.target = (Entity) arraylist.get(this.rand.nextInt(arraylist.size()));
		} else {
			while (true) {
				this.targetX = 0.0D;
				this.targetY = (double) (70.0F + this.rand.nextFloat() * 50.0F);
				this.targetZ = 0.0D;
				this.targetX += (double) (this.rand.nextFloat() * 120.0F - 60.0F);
				this.targetZ += (double) (this.rand.nextFloat() * 120.0F - 60.0F);
				double d0 = this.posX - this.targetX;
				double d1 = this.posY - this.targetY;
				double d2 = this.posZ - this.targetZ;
				boolean flag = d0 * d0 + d1 * d1 + d2 * d2 > 100.0D;
				if (flag) {
					break;
				}
			}

			this.target = null;
		}

	}

	/**
	 * + Simplifies the value of a number by adding/subtracting 180 to the point
	 * that the number is between -180 and 180.
	 */
	private float simplifyAngle(double parDouble1) {
		return (float) MathHelper.wrapAngleTo180_double(parDouble1);
	}

	/**
	 * + Updates the state of the enderdragon's current endercrystal.
	 */
	private void updateDragonEnderCrystal() {
		if (this.healingEnderCrystal != null) {
			if (this.healingEnderCrystal.isDead) {
				if (!this.worldObj.isRemote) {
					this.attackEntityFromPart(this.dragonPartHead, DamageSource.setExplosionSource((Explosion) null),
							10.0F);
				}

				this.healingEnderCrystal = null;
			} else if (this.ticksExisted % 10 == 0 && this.getHealth() < this.getMaxHealth()) {
				this.setHealth(this.getHealth() + 1.0F);
			}
		}

		if (this.rand.nextInt(10) == 0) {
			float f = 32.0F;
			List<EntityEnderCrystal> list = this.worldObj.getEntitiesWithinAABB(EntityEnderCrystal.class,
					this.getEntityBoundingBox().expand((double) f, (double) f, (double) f));
			EntityEnderCrystal entityendercrystal = null;
			double d0 = Double.MAX_VALUE;

			for (int i = 0, l = list.size(); i < l; ++i) {
				EntityEnderCrystal entityendercrystal1 = list.get(i);
				double d1 = entityendercrystal1.getDistanceSqToEntity(this);
				if (d1 < d0) {
					d0 = d1;
					entityendercrystal = entityendercrystal1;
				}
			}

			this.healingEnderCrystal = entityendercrystal;
		}

	}
}
