/*
 * Copyright (C) 2011 The Guava Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.common.base;

import static com.google.common.base.Preconditions.checkNotNull;

import java.util.Collections;
import java.util.Set;

import javax.annotation.Nullable;

import com.google.common.annotations.GwtCompatible;

/**
 * Implementation of an {@link Optional} containing a reference.
 */
@GwtCompatible
final class Present<T> extends Optional<T> {
	private static final long serialVersionUID = 0;

	private final T reference;

	Present(T reference) {
		this.reference = reference;
	}

	@Override
	public Set<T> asSet() {
		return Collections.singleton(reference);
	}

	@Override
	public boolean equals(@Nullable Object object) {
		if (object instanceof Present) {
			Present<?> other = (Present<?>) object;
			return reference.equals(other.reference);
		}
		return false;
	}

	@Override
	public T get() {
		return reference;
	}

	@Override
	public int hashCode() {
		return 0x598df91c + reference.hashCode();
	}

	@Override
	public boolean isPresent() {
		return true;
	}

	@Override
	public Optional<T> or(Optional<? extends T> secondChoice) {
		checkNotNull(secondChoice);
		return this;
	}

	@Override
	public T or(Supplier<? extends T> supplier) {
		checkNotNull(supplier);
		return reference;
	}

	@Override
	public T or(T defaultValue) {
		checkNotNull(defaultValue, "use Optional.orNull() instead of Optional.or(null)");
		return reference;
	}

	@Override
	public T orNull() {
		return reference;
	}

	@Override
	public String toString() {
		return "Optional.of(" + reference + ")";
	}

	@Override
	public <V> Optional<V> transform(Function<? super T, V> function) {
		return new Present<V>(checkNotNull(function.apply(reference),
				"the Function passed to Optional.transform() must not return null."));
	}
}
