/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.commons.lang3.text.translate;

import java.io.IOException;
import java.io.Writer;

/**
 * Translates escaped Unicode values of the form \\u+\d\d\d\d back to Unicode.
 * It supports multiple 'u' characters and will work with or without the +.
 *
 * @since 3.0
 * @version $Id: UnicodeUnescaper.java 1436770 2013-01-22 07:09:45Z ggregory $
 */
public class UnicodeUnescaper extends CharSequenceTranslator {

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int translate(final CharSequence input, final int index, final Writer out) throws IOException {
		if (input.charAt(index) == '\\' && index + 1 < input.length() && input.charAt(index + 1) == 'u') {
			// consume optional additional 'u' chars
			int i = 2;
			while (index + i < input.length() && input.charAt(index + i) == 'u') {
				i++;
			}

			if (index + i < input.length() && input.charAt(index + i) == '+') {
				i++;
			}

			if (index + i + 4 <= input.length()) {
				// Get 4 hex digits
				final CharSequence unicode = input.subSequence(index + i, index + i + 4);

				try {
					final int value = Integer.parseInt(unicode.toString(), 16);
					out.write((char) value);
				} catch (final NumberFormatException nfe) {
					throw new IllegalArgumentException("Unable to parse unicode value: " + unicode, nfe);
				}
				return i + 4;
			} else {
				throw new IllegalArgumentException("Less than 4 hex digits in unicode value: '"
						+ input.subSequence(index, input.length()) + "' due to end of CharSequence");
			}
		}
		return 0;
	}
}
